/******************************************************************************/
/* 
/*	This class serves as reference implementation for Teledrive(C) 
	hardware access. You can use this code,	modify or include it in your 
	application. The code is widely commented and may in some minor points not
	conform to carapi specifications since sw-development evolves. 

							IAV-GmbH
							Carnotstrae 1
							10587 Berlin
							Germany
							Phone: +49 (30) 39978-0
							http://www.iav.de

	Included main features (others might be added after this was written):
	- complete sound functionality like volume, fader, source select...
	- Power management
	- CAN communication
	- K-Line communication
	- Tuner (AM, FM, Scanning)
	- Rotary switch
	  
	Hardware access is done message based. All messages	are encoded and 
	decoded by carapi.dll.
	
	The very basic principle is: 
	1. connect to carapi.dll at startup of application
	2. encode a message
	3. send it to carapi.dll
	4. receive an answer from carapi.dll (callback)
	5. decode the answer and process it
	6. disconnect from carapi.dll when exiting application
	
	It is also possible to receive a message without prior sending.
	That may be caused by an hardware event like e.g. 'ignition-on'.

	The more detailed description of the communication cares for hardware 
	restrictions. Lets see an example: Given you want to set the volume
	to a certain value, you first have to know about the minimum and 
	maximum value accepted by hardware. So first you have to 
	query the volume capabilities (range) second you send a 'set volume' 
	command. Once you know the minimum and maximum you dont have to query 
	again when setting a new volume level.
	For more information please refer to carapi specification and code below.
	
	Important note:
	Teledrive(C) is closely related to its automotive environment. Hardware
	power management ensures proper functionality in 'worst-case' 
	situations (low battery, software failure). When the system is in
	'normal mode' at least one PC-application or PC-service is responsible to 
	interact to the hardware power management (send startup messages, react 
	on shutdown messages).
	Teledrive(C) will not behave as expected and data loss may occur in your 
	application if this is ignored!	Please see carapi specification. 

	related files:
	carapi.lib		carapi interface library
	carapi.h		carapi interface definition
/*	carmsg.h		carapi message definition
/*
/******************************************************************************/


#include "stdafx.h"

#include "Communication.h"
#include <assert.h>

#ifdef __cplusplus
extern "C" 
{
	#include "carapi.h"
}
#endif	// __cplusplus

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/******************************************************************************/
/*
/*	Globals			
/*		
/******************************************************************************/
void __stdcall OnReceiveMessage(PVOID a, DWORD b, DWORD c , 
								DWORD d, DWORD e, PBYTE f);

void RedirectDialRequest(Communication* p, int direction, int ticks);

DataLock		g_syncLock;
AreaDescription g_areaDescr[4];
Communication	*pComm				= NULL;

FN_CALLBACK		g_pCallback			= NULL;
AUDIOCAPS		*g_pm_AudioCaps		= NULL;
CANINFO			*g_pm_CANInfo		= NULL;
KLINEINFO		*g_pm_KLineInfo[2];
FREQPARAMS		g_freqParams;
FREQULIST		g_scanList;

int				g_dialTarget		= TARGETNONE;
int				g_dialValue			= 0;
int				g_tickMultiplier	= 1;
int				g_scanListEntry		= 0;

/******************************************************************************/
/*
/*  Funktion:		Constructor
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:				
/*		
/******************************************************************************/
Communication::Communication(FN_CALLBACK fnName)
{	
	//Pointer auf Funktion ?
	assert( fnName != NULL );

	m_Callback			= fnName;
	m_CurrentPrio		= 2;
	m_connected			= false;
	
	
	g_dialTarget		= TARGETNONE;
	g_dialValue			= 0;
	g_tickMultiplier	= 1;
	g_pCallback			= fnName;
	pComm				= this;

	memset(&m_AudioCaps, 0x00, sizeof(m_AudioCaps));
	memset(&g_freqParams, 0x00, sizeof(g_freqParams));
	memset(&m_CANInfo, 0x00, sizeof(m_CANInfo));
	memset(&m_KLineInfo, 0x00, sizeof(m_KLineInfo));
	
	m_CANInfo.connectState = DISCONNECTED;
	m_KLineInfo[0].connectState = DISCONNECTED;
	m_KLineInfo[1].connectState = DISCONNECTED;

	g_pm_AudioCaps		= &m_AudioCaps;
	g_pm_CANInfo		= &m_CANInfo;
	g_pm_KLineInfo[0]	= &m_KLineInfo[0];
	g_pm_KLineInfo[1]	= &m_KLineInfo[1];

	DBGTRACE(("Constructor executed"));
}


/******************************************************************************/
/*
/*  Funktion:		Destructor
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:				
/*		
/******************************************************************************/
Communication::~Communication()
{	
	g_pm_AudioCaps		= NULL;
	g_pm_KLineInfo[0]	= NULL;
	g_pm_KLineInfo[1]	= NULL;
	g_pm_CANInfo		= NULL;
	pComm				= NULL;

	if ( m_connected ){ 
		DisConnect();
		m_connected = false;
		DBGTRACE(("Destructor executed disconnect."));
	}
	
	DBGTRACE(("Destructor succeeded."));
	return;
}


/******************************************************************************/
/*
/*  Funktion:		Connect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		At application startup we need to connect to 
/*					carapi.dll by registering a callback. Messages 
/*					can only be sent and received after the callback 
/*					was registered successful.		
/*		
/******************************************************************************/
bool Communication::Connect(void)
{	
	//try to register a callback for receiving messages
	if ( !carapiRecvRegisterCallback(&m_linkId, NULL, OnReceiveMessage) ){
		(*m_Callback)(ERRORDLL, NULL);
		m_connected = false;
		DBGTRACE(("Connect failed. Dll rejects to RegisterCallback"));
		return false;
	}
	else{
		(*m_Callback)(DLLCONNECT, m_linkId);
		m_connected = true;
		DBGTRACE(("Connect succeeded with LinkID=%d", m_linkId));
		

		//tell the Power-Management that we are awake now
		SendPCStartUpMessage();
		return true;
	}
}


/******************************************************************************/
/*
/*  Funktion:		DisConnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		When application terminates we unregister the call-
/*					back. Then we are disconnected. No further messages
/*					can be sent or received. 		
/*		
/******************************************************************************/
bool Communication::DisConnect(void)
{
	unsigned char status = carapiRecvUnregisterCallback(m_linkId);

	if ( 0 == status ){
		DBGTRACE(("DisConnect - error."));
		return false;
	}
	else{
		m_connected = false;
		DBGTRACE(("DisConnect - succeeded."));
		return true;
	}
}


/******************************************************************************/
/*
/*  Funktion:		OnReceiveMessage
/*	
/*
/*	Parameters:		a:	dont care
/*					b:	dont care
/*					c:	dont care
/*					d:	carapi message id
/*					e:	error flag: 0 means 'no error'
/*					f:	Pointer to carapi message
/*
/*	Return Value:	--
/*
/*	Comments:		All messages received by carapi.dll end up here. If a msg
/*					returned with error (e != 0) decoding may fail because
/*					the contents may be invalid!	
/*		
/******************************************************************************/
void __stdcall OnReceiveMessage(PVOID a, DWORD b, DWORD c , 
								DWORD d, DWORD e, PBYTE f)
{
	DWORD	status	 = 0;
	int		value	 = 0;

	int		val1	= 0;
	int		val2	= 0;

	PBYTE	pBlock	 = NULL;
	DWORD	blockLen = 0;
	DWORD	blockTyp = 0;

	DWORD	bandListLen	= 0;
	DWORD	areaDescLen = 0;
	DWORD	bandDescLen = 0;
	DWORD	areaListLen = 0;

	PBYTE	pAreaList  = NULL;
	PBYTE	pg_areaDescr = NULL;
	PBYTE	pBandList  = NULL;
	PBYTE	pBandDescr = NULL;
	PBYTE	pArList	   = NULL;

	int	dummyArCode	= 0;
	int fMin = 0;
	int fMax = 0;
	int deltaF = 0;
	int dummyBand = 0;
	
	DBGTRACE(("Callback entered."));

	//Pointer to message valid?
	assert( f != NULL );
	
	if ( !g_pCallback ){
		DBGTRACE(("Callback entry condition false."));
		return;
	}
	
	int i = 0;

	//Array-Index bestimmen
	if ( b == KLINEMODUL1 ){
		DBGTRACE(("Callback - Message from K-Line 1"));
		i = 0;
	}
	else if ( b == KLINEMODUL2 ) {
		DBGTRACE(("Callback - Message from K-Line 2"));
		i = 1;
	}

	//Error flag set?
	if ( e ){

		//some modules need error notification, no decoding here
		switch ( d ){
			case CARMSG_AudioTuner_ScanBand_CON:
			g_scanListEntry = 0;
			DBGTRACE(("Callback ScanBand_CON Scan error."));
			(*g_pCallback)(SCANERROR, 0);
			return;

			case CARMSG_IOCAN_PhysicalConnect_CON:
			case CARMSG_IOCAN_LogicalConnect_CON:
				DBGTRACE(("Callback CAN_XXX_Connect_CON error."));
				g_syncLock.Aquire();
				g_pm_CANInfo->connectState = DISCONNECTED;
				g_syncLock.Release();
				(*g_pCallback)(CONNECTERROR, (LPARAM)g_pm_CANInfo);
				return;

			case CARMSG_IOKLINE_PhysicalConnect_CON:
			case CARMSG_IOKLINE_LogicalConnect_CON:
				DBGTRACE(("Callback KLine-Connect error."));
				g_syncLock.Aquire();
				g_pm_KLineInfo[i]->connectState = DISCONNECTED;
				g_syncLock.Release();
				(*g_pCallback)(KCONNECTERROR, e);
				return;
				
			case CARMSG_IOKLINE_PhysicalDisconnect_CON:
			case CARMSG_IOKLINE_LogicalDisconnect_CON:
				(*g_pCallback)(KDISCONERROR, e);
				return;

			default:
				DBGTRACE(("Callback errors on msgId 0x%x not handled.", d));
				break;
		}

		DBGTRACE(("Callback error flag of Msg 0x%x was set to 0x%x. Giving up...", d, e));
		(*g_pCallback)(ERRGENERAL, (int)e);
		return;	
	}
	
	//now lets see what message we have
	switch ( d ){
	
	case CARMSG_AudioSelector_QueryCaps_CON:
		
		//decode a value that is in this message
		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Selector-Con, value %d", value));
		}
		else{
			DBGTRACE(("Callback Decf failed on Selector-Con"));
			//not yet implemented
		}
		break;

	case CARMSG_AudioVolume_QueryCaps_CON:

		g_syncLock.Aquire();

		//decode two values included in this message
		if ( carapiBlockDecf(f, "vv", &g_pm_AudioCaps->volMin, &g_pm_AudioCaps->volMax) ){
			DBGTRACE(("Callback Decf Volume-Con, min %d", g_pm_AudioCaps->volMin));
			DBGTRACE(("Callback Decf Volume-Con, max %d", g_pm_AudioCaps->volMax));
			//pass result to related module
			(*g_pCallback)(VOLCAPS, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on Volume-Con"));
			//pass error to related module
			(*g_pCallback)(AUDERROR, value);
		}

		g_syncLock.Release();
		break;
	
	case CARMSG_AudioBalance_QueryCaps_CON:

		g_syncLock.Aquire();
		if ( carapiBlockDecf(f, "vvv", &g_pm_AudioCaps->balLeft, 
			&g_pm_AudioCaps->balRight, &g_pm_AudioCaps->balCntr) )
		{
			DBGTRACE(("Callback Decf Bal-Con, left %d", g_pm_AudioCaps->balLeft));
			DBGTRACE(("Callback Decf Bal-Con, right %d", g_pm_AudioCaps->balRight));
			DBGTRACE(("Callback Decf Bal-Con, cntr %d", g_pm_AudioCaps->balCntr));
			(*g_pCallback)(BALCAPS, value);
			
		}
		else{
			DBGTRACE(("Callback Decf failed on Bal-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		g_syncLock.Release();
		break;
	
	case CARMSG_AudioFader_QueryCaps_CON:

		g_syncLock.Aquire();
		if ( carapiBlockDecf(f, "vvv", &g_pm_AudioCaps->fadFront, 
			&g_pm_AudioCaps->fadRear, &g_pm_AudioCaps->fadCntr) )
		{	
			DBGTRACE(("Callback Decf Fad-Con, front %d", g_pm_AudioCaps->fadFront));
			DBGTRACE(("Callback Decf Fad-Con, rear %d", g_pm_AudioCaps->fadRear));
			DBGTRACE(("Callback Decf Fad-Con, cntr %d", g_pm_AudioCaps->fadCntr));
			(*g_pCallback)(FADCAPS, value);	
			
		}
		else{
			DBGTRACE(("Callback Decf failed on Fad-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		g_syncLock.Release();
		break;

	case CARMSG_AudioFilterBass_QueryCaps_CON:

		g_syncLock.Aquire();
		if ( carapiBlockDecf(f, "vvvv", &g_pm_AudioCaps->basMin, 
			&g_pm_AudioCaps->basMax, &g_pm_AudioCaps->basCntr, &g_pm_AudioCaps->basCntrFreq) )
		{
			DBGTRACE(("Callback Decf Bas-Con, min %d", g_pm_AudioCaps->basMin));
			DBGTRACE(("Callback Decf Bas-Con, max %d", g_pm_AudioCaps->basMax));
			DBGTRACE(("Callback Decf Bas-Con, cntr %d", g_pm_AudioCaps->basCntr));
			DBGTRACE(("Callback Decf Bas-Con, cntrF %d", g_pm_AudioCaps->basCntrFreq));
			(*g_pCallback)(BASCAPS, value);	
		}
		else{
			DBGTRACE(("Callback Decf failed on Bas-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		g_syncLock.Release();
		break;

	case CARMSG_AudioFilterTreble_QueryCaps_CON:

		g_syncLock.Aquire();
		if ( carapiBlockDecf(f, "vvvv", &g_pm_AudioCaps->trebMin, 
			&g_pm_AudioCaps->trebMax, &g_pm_AudioCaps->trebCntr, &g_pm_AudioCaps->trebCntrFreq) )
		{	
			DBGTRACE(("Callback Decf Treb-Con, min %d", g_pm_AudioCaps->trebMin));
			DBGTRACE(("Callback Decf Treb-Con, max %d", g_pm_AudioCaps->trebMax));
			DBGTRACE(("Callback Decf Treb-Con, cntr %d", g_pm_AudioCaps->trebCntr));
			DBGTRACE(("Callback Decf Treb-Con, cntrF %d", g_pm_AudioCaps->trebCntrFreq));
			(*g_pCallback)(TREBCAPS, value);	
			
		}
		else{
			DBGTRACE(("Callback Decf failed on Treb-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		g_syncLock.Release();
		break;

	case CARMSG_AudioFilterLoudness_QueryCaps_CON:

		g_syncLock.Aquire();
		if ( carapiBlockDecf(f, "vvv", &g_pm_AudioCaps->loudMin, 
			&g_pm_AudioCaps->loudMax, &g_pm_AudioCaps->loudCntr) )
		{
			DBGTRACE(("Callback Decf Loud-Con, min %d", g_pm_AudioCaps->loudMin));
			DBGTRACE(("Callback Decf Loud-Con, max %d", g_pm_AudioCaps->loudMax));
			DBGTRACE(("Callback Decf Loud-Con, cntr %d", g_pm_AudioCaps->loudCntr));
			(*g_pCallback)(LOUDCAPS, value);	
			
		}
		else{
			DBGTRACE(("Callback Decf failed on Loud-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		g_syncLock.Release();
		break;

	case CARMSG_AudioVolume_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf VolPos-Con, value: %d", value));
			(*g_pCallback)(VOLCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on VolPos-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioSelector_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Selector-Con, value: %d", value));
			(*g_pCallback)(SRCCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on Selector-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioBalance_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Bal-Con, value: %d", value));
			(*g_pCallback)(BALCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on Bal-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioFader_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Fad-Con, value: %d", value));
			(*g_pCallback)(FADCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on Fad-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioFilterBass_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Bas-Con, value: %d", value));
			(*g_pCallback)(BASCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on Bas-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioFilterTreble_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Treb-Con, value: %d", value));
			(*g_pCallback)(TREBCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on Treb-Con"));
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioFilterLoudness_Position_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf Loudness-Con, value: %d", value));
			(*g_pCallback)(LOUDCON, value);
		}
		else{
			(*g_pCallback)(AUDERROR, value);
		}
		break;

	case CARMSG_AudioTuner_QueryCaps_CON:
		pBlock	 = NULL;
		blockLen = 0;
		blockTyp = 0;

		areaListLen = 0;
		bandListLen	= 0;
		areaDescLen = 0;
		bandDescLen = 0;
		
		dummyArCode	= 0;
		fMin		= 0;
		fMax		= 0;
		dummyBand	= 0;
	
		pAreaList	= NULL;
		pg_areaDescr = NULL;
		pBandList	= NULL;
		pBandDescr	= NULL;

		DBGTRACE(("Callback TunCapsCon start decoding."));
		
		//Get pointer to area-Description list 
		carapiBlockDecBlock(f, &pArList, NULL, NULL);
		assert(pArList != NULL);

		//Get pointer and length of first element in area-Description
		if ( carapiBlockDecBlock(pArList, &pg_areaDescr, &areaListLen, NULL) ){
			assert( pg_areaDescr != NULL );
			assert( areaListLen != 0 );
			
			int area = 0; //Europe, USA, ...
			int bandType = 0; //FM, AM, ...

			g_syncLock.Aquire();

			//iterate elements in area-Description list
			while( areaListLen > 0 ){
				
				carapiBlockGetLength(pg_areaDescr, &areaDescLen);
				assert( areaDescLen != 0 );
				
				//Get id of area and pointer to band list
				carapiBlockDecf(pg_areaDescr, "vs",  &area, &pBandList);
				assert( pBandList != NULL );
				assert( area > 0 );
				area--;

				DBGTRACE(("Callback TunCapsCon Decf 1 ArCode value %d", area));
				
				if ( carapiBlockDecBlock(pBandList, &pBandDescr, &bandListLen, NULL) ){
					assert( pBandDescr != NULL );
					assert( bandListLen != 0 );
					
					//iterate elements in band-list in current area-description element 
					while ( bandListLen > 0 ){
						
						//Get length of current band-list element
						carapiBlockGetLength(pBandDescr, &bandDescLen);
						assert( bandDescLen != 0 );
						
						//get values of band-list element
						carapiBlockDecf(pBandDescr, "vvvv",  &bandType, &fMin, &fMax, &deltaF);
						assert( bandType > 0 );
						bandType--;

						//save values
						g_areaDescr[area].areaCode = (area + 1);
						g_areaDescr[area].band[bandType].bandType = (bandType+1);
						g_areaDescr[area].band[bandType].minFreq = fMin;
						g_areaDescr[area].band[bandType].maxFreq = fMax;
						g_areaDescr[area].band[bandType].deltaF  = deltaF;
					    
						DBGTRACE(("Callback TunCapsCon Decf 2 Band value %d", bandType));
						DBGTRACE(("Callback TunCapsCon Decf 2 fMin value %d", fMin));
						DBGTRACE(("Callback TunCapsCon Decf 2 fMax value %d", fMax));
						DBGTRACE(("Callback TunCapsCon Decf 2 deltaF value %d", deltaF));
						
						//go to next band list element
						pBandDescr += bandDescLen;
						bandListLen -= bandDescLen;
						assert( bandDescLen >  0);
					}
				}
				else{
					DBGTRACE(("Callback DecBlock 2 failed on TunerCaps-Con"));
				}
				
				//go to next area list element
				pg_areaDescr += areaDescLen;
				areaListLen -= areaDescLen;
				assert( areaDescLen > 0 );
			}

			g_syncLock.Release();
			DBGTRACE(("Callback TunCapsCon finished decoding."));

			//call related module
			(*g_pCallback)(TUNCAPS, value);

		}
		else{
			DBGTRACE(("Callback DecBlock 1 failed on TunerCaps-Con"));
			(*g_pCallback)(TUNERROR, value);
		}
		break;
	
	case CARMSG_AudioTuner_SetFrequency_CON:
		
		g_syncLock.Aquire();
	
		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf SetFrequency_CON, value: %d", value));
			g_freqParams.frequency = value;
			(*g_pCallback)(FREQCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on SetFrequency_CON"));
			(*g_pCallback)(TUNERROR, value);
		}
		g_syncLock.Release();
		break;
	
	case CARMSG_AudioTuner_SelectBand_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf SelectBand_CON, value: %d", value));
			(*g_pCallback)(BANDCON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on SelectBand_CON"));
			(*g_pCallback)(TUNERROR, value);
		}
		break;

	case CARMSG_AudioTuner_SelectArea_CON:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Decf SelectArea_CON, value: %d", value));
			(*g_pCallback)(AREACON, value);
		}
		else{
			DBGTRACE(("Callback Decf failed on SelectArea_CON"));
			(*g_pCallback)(TUNERROR, value);
		}
		break;

	case CARMSG_AudioTuner_ScanBand_EVT:{
		
		PBYTE	pFreqDesc	= NULL;
		PBYTE   pItem		= NULL;
		DWORD	freqDescLen = 0;
		DWORD	itemLen		= 0;
		int		flags		= 0;

		DBGTRACE(("Callback ScanBand_EVT start decoding."));
		
		//get pointer to freqency-Description list
		carapiBlockDecBlock(f, &pFreqDesc, NULL, NULL);
		assert(pFreqDesc != NULL);
	
		//get pointer to first element of Freqency-Description list
		if ( carapiBlockDecBlock(pFreqDesc, &pItem, &freqDescLen, NULL) ){
			assert( pItem != NULL );
			assert( freqDescLen != 0 );

			//get length of element
			carapiBlockGetLength(pItem, &itemLen);
			assert(itemLen > 0);
			
			g_syncLock.Aquire();

			//iterate frequency-Description-list elements
			while( freqDescLen > 0 ){
				
				//range check
				assert( (g_scanListEntry > -1) && (g_scanListEntry <= (FLIST_ITEMS-1)) );

				carapiBlockDecf(pItem, "vvv", &g_scanList.item[g_scanListEntry].frequency, 
											  &g_scanList.item[g_scanListEntry].signalQuality,
											  &flags);
				
				//Stereo-Bit
				g_scanList.item[g_scanListEntry].stereo = (flags & 0x00000001);
				
				//Rds-Bit
				flags &= 0x00000010;
				flags >>= 1;
				g_scanList.item[g_scanListEntry].rds = flags;
				
				//go to next element in frequency-Description-list
				pItem += itemLen;
				freqDescLen -= itemLen;
				g_scanListEntry++;
			}
			
			//save decoded values
			g_scanList.nEntries = g_scanListEntry;

			DBGTRACE(("Callback ScanBand_EVT decoded %d items", g_scanListEntry));
			DBGTRACE(("Callback ScanBand_EVT first item: f:%d  Qlty:%d  St:%d  Rds:%d", 
				g_scanList.item[0].frequency,
				g_scanList.item[0].signalQuality,
				g_scanList.item[0].stereo,
				g_scanList.item[0].rds));
			DBGTRACE(("Callback ScanBand_EVT last item: f:%d  Qlty:%d  St:%d  Rds:%d", 
				g_scanList.item[(g_scanListEntry-1)].frequency,
				g_scanList.item[(g_scanListEntry-1)].signalQuality,
				g_scanList.item[(g_scanListEntry-1)].stereo,
				g_scanList.item[(g_scanListEntry-1)].rds));

			g_syncLock.Release();

			//call related module
			(*g_pCallback)(SCANRUN, value);
		}
		else{
			DBGTRACE(("Callback ScanBand_EVT failed on BlockDecBlock"));
		}
		break;
	}

	case CARMSG_AudioTuner_ScanBand_CON:
		g_scanListEntry = 0;
		DBGTRACE(("Callback ScanBand_CON Scan finished."));
		(*g_pCallback)(SCANCOMPL, value);
		break;

	case CARMSG_Dial_TurnRight_EVT:
		
		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Dial right-Event"));
			g_syncLock.Aquire();

			if ( g_dialTarget != TARGETNONE ){
				RedirectDialRequest(pComm, DIALRIGHT, value);
			}
			else{
				(*g_pCallback)(DIALRIGHT, value);
			}
			g_syncLock.Release();
		}
		else{
			(*g_pCallback)(DIALERROR, value);	
		}
		break;
	
	case CARMSG_Dial_TurnLeft_EVT:

		if ( carapiBlockDecf(f, "v", &value) ){
			DBGTRACE(("Callback Dial left-Event"));
			g_syncLock.Aquire();
			if ( g_dialTarget != TARGETNONE ){
				RedirectDialRequest(pComm, DIALLEFT, value);
			}
			else{
				(*g_pCallback)(DIALLEFT, value);
			}
			g_syncLock.Release();
		}
		else{
			(*g_pCallback)(DIALERROR, value);
		}
		break;

	case CARMSG_Key_Down_EVT:
		DBGTRACE(("Callback Dial down-Event"));
		(*g_pCallback)(DIALPUSH, value);
		break;

	case CARMSG_Key_Up_EVT:
		DBGTRACE(("Callback Dial up-Event"));
		(*g_pCallback)(DIALRELEASE, value);
		break;
	
	case CARMSG_IOCAN_PhysicalConnect_EVT:
		DBGTRACE(("OnReceiveMessage - PhysicalConnect_EVT received"));

		g_syncLock.Aquire();
		carapiBlockDecf(f, "v", &g_pm_CANInfo->physicalId);
		g_pm_CANInfo->connectState = PHYSICAL_RDY;
		g_syncLock.Release();
		(*g_pCallback)(PHYSCONNECT, g_pm_CANInfo->physicalId);
		
		DBGTRACE(("OnReceiveMessage - PhysicalConnect_EVT physId: %d", 
					g_pm_CANInfo->physicalId));
		break;
	
	case CARMSG_IOCAN_LogicalConnect_EVT:
		DBGTRACE(("OnReceiveMessage - LogicalConnect_EVT received"));
		
		g_syncLock.Aquire();
		if ( carapiBlockDecf(f, "vv", &g_pm_CANInfo->physicalId, &g_pm_CANInfo->logicalId) ){
			DBGTRACE(("OnReceiveMessage - physCanConnect: %d", g_pm_CANInfo->physicalId));
			DBGTRACE(("OnReceiveMessage - logCanConnect: %d", g_pm_CANInfo->logicalId));
		}
		g_pm_CANInfo->connectState = LOGICAL_RDY;
		g_syncLock.Release();
		(*g_pCallback)(LOGCONNECT, (LPARAM)g_pm_CANInfo);
		break;
	
	case CARMSG_IOCAN_PhysicalConnect_CON:
		DBGTRACE(("OnReceiveMessage - PhysicalConnect_CON received"));
		break;
	
	case CARMSG_IOCAN_PhysicalDisconnect_CON:
		DBGTRACE(("OnReceiveMessage - PhysicalDisconnect_CON received"));
		break;

	case CARMSG_IOCAN_PhysicalDisconnect_EVT:
		DBGTRACE(("OnReceiveMessage - PhysicalDisconnect_EVT received"));
		
		g_syncLock.Aquire();
		g_pm_CANInfo->connectState = DISCONNECTED;
		g_syncLock.Release();
		(*g_pCallback)(PHYSDISCONNECT, (LPARAM)g_pm_CANInfo);
		break;
	
	case CARMSG_IOCAN_LogicalDisconnect_EVT:
		DBGTRACE(("OnReceiveMessage - LogicalDisconnect_EVT received"));

		g_syncLock.Aquire();
		g_pm_CANInfo->connectState = PHYSICAL_RDY;
		g_syncLock.Release();
		(*g_pCallback)(LOGDISCONNECT, g_pm_CANInfo->physicalId);
		break;

	case CARMSG_IOCAN_LogicalConnect_CON:
		DBGTRACE(("OnReceiveMessage - LogicalConnect_CON received"));
		
		g_syncLock.Aquire();
		carapiBlockDecf(f, "vv", &g_pm_CANInfo->physicalId, &g_pm_CANInfo->logicalId);
		g_syncLock.Release();

		DBGTRACE(("OnReceiveMessage - LogicalConnect_CON physId: %d", g_pm_CANInfo->physicalId));
		DBGTRACE(("OnReceiveMessage - LogicalConnect_CON logId: %d", g_pm_CANInfo->logicalId));
		break;

	case CARMSG_IOCAN_LogicalDisconnect_CON:
		DBGTRACE(("OnReceiveMessage - LogicalDisonnect_CON received"));
		
		g_syncLock.Aquire();
		carapiBlockDecf(f, "vv", &g_pm_CANInfo->physicalId, &g_pm_CANInfo->logicalId);
		g_syncLock.Release();

		DBGTRACE(("OnReceiveMessage - LogicalDisonnect_CON physId: %d", g_pm_CANInfo->physicalId));
		DBGTRACE(("OnReceiveMessage - LogicalDisonnect_CON logId: %d", g_pm_CANInfo->logicalId));
		break;

	case CARMSG_IOCAN_Send_CON:
		DBGTRACE(("OnReceiveMessage - Send_CON received"));
		(*g_pCallback)(TXDATA, 0);
		break;

	case CARMSG_IOCAN_Recv_EVT:{
		UINT	physId;
		UINT	logId;
		UCHAR	*pCANMsg;
		UCHAR	*pList;
		ULONG	listLen;
		ULONG	itemLen;
		UCHAR	*pListItem;
		CANDATA canItem;

		DBGTRACE(("OnReceiveMessage - Recv_EVT received"));

		carapiBlockDecf(f, "vvs",  &physId, &logId, &pList);
		
		DBGTRACE(("OnReceiveMessage - physId: %d", physId));
		DBGTRACE(("OnReceiveMessage - logId: %d", logId));
		
		//a list of CAN messages was received - get pointer and list length
		if ( !carapiBlockDecBlock(pList, &pListItem, &listLen, NULL) ){
			assert(false);
			return;
		}

		DBGTRACE(("OnReceiveMessage - listLen: %d", listLen));
		
		// iterate list of CAN messages
		while ( listLen > 0 ){
			carapiBlockGetLength(pListItem, &itemLen);
			DBGTRACE(("OnReceiveMessage - itemLen: %d", itemLen));
			
			//decode CAN-Identifier, flags and CAN-message
			carapiBlockDecf(pListItem, "vvu", &canItem.canId, 
											  &canItem.rxFlags, 
											  &pCANMsg, 
											  &canItem.msgLen);
		
			
			
			DBGTRACE(("OnReceiveMessage - CAN Id: 0x%x", canItem.canId));
			DBGTRACE(("OnReceiveMessage - CAN rxFlags: 0x%x", canItem.rxFlags));
			DBGTRACE(("OnReceiveMessage - CAN msgLen: %d", canItem.msgLen));
			memcpy(&canItem.msg, pCANMsg, canItem.msgLen);

#ifdef _DEBUG
			TCHAR stream[64] = {0};
			for (unsigned int i = 0; i < canItem.msgLen; i++){
				sprintf(stream + + strlen(stream), "%02x ", canItem.msg[i]);
			}
			DBGTRACE(("OnReceiveMessage - CAN Msg: %s", stream));
#endif
			(*g_pCallback)(RXDATA, (LPARAM)&canItem);	

			//set pointer to next element in list
			assert(itemLen > 0);
			pListItem += itemLen;
			listLen -= itemLen;
			DBGTRACE(("OnReceiveMessage - new listLen: %d", listLen));
		}
		
		DBGTRACE(("OnReceiveMessage - Recv_EVT received"));
		break;
	}
	
	case CARMSG_IOCAN_LogicalStatus_EVT:{
		UINT	physId;
		UINT	logId;
		//UINT	busStatus;
		//UINT	rxErrCnt;
		//UINT	txErrCnt;
		//UINT	errCode;
		UCHAR	*pList;
		UCHAR	*pListItem;
		ULONG	listLen;
		ULONG	itemLen;
		CANSTATUS canStat;

		carapiBlockDecf(f, "vvs",  &physId, &logId, &pList);
		DBGTRACE(("OnReceiveMessage - CARMSG_IOCAN_LogicalStatus_EVT"));

		if ( !carapiBlockDecBlock(pList, &pListItem, &listLen, NULL) ){
			assert(false);
			return;
		}

		while ( listLen > 0 ){
			carapiBlockGetLength(pListItem, &itemLen);
			DBGTRACE(("OnReceiveMessage - itemLen: %d", itemLen));

			carapiBlockDecf(pListItem, "vvvv", &canStat.CANbusState, 
											  &canStat.txErrorCount, 
											  &canStat.rxErrorCount, 
											  &canStat.errorDescriptor);
		
			
			
			DBGTRACE(("OnReceiveMessage - BusStatus: 0x%x", canStat.CANbusState));
			DBGTRACE(("OnReceiveMessage - TXERR:  0x%x", canStat.txErrorCount));
			DBGTRACE(("OnReceiveMessage - RXERR:  0x%x", canStat.rxErrorCount));
			DBGTRACE(("OnReceiveMessage - DESCR:  0x%x", canStat.errorDescriptor));

			(*g_pCallback)(STATUSMSG, (LPARAM)&canStat);	

			assert(itemLen > 0);
			pListItem += itemLen;
			listLen -= itemLen;
		}
		break;
	}

	case CARMSG_PowerSupply_SwitchOff_CON:
		DBGTRACE(("OnReceiveMessage - PowerSupply_SwitchOff confirmation"));
		(*g_pCallback)(SHUTDOWNCON, 0);
		break;
	
	case CARMSG_PowerSupply_SwitchOff_EVT:
		carapiBlockDecf(f, "v",  &value);
		DBGTRACE(("OnReceiveMessage - PowerSupply_SwitchOff_EVT Time: %d ms", value));
		(*g_pCallback)(SHUTDOWNEVT, value);
		break;
	
	case CARMSG_Ignition_State_CON:
		carapiBlockDecf(f, "v",  &value);
		DBGTRACE(("OnReceiveMessage - Ignition State is %d", value));
		(*g_pCallback)(IGNITIONCON, value);
		break;

	case CARMSG_Ignition_State_EVT:
		carapiBlockDecf(f, "v",  &value);
		DBGTRACE(("OnReceiveMessage - Ignition State is %d", value));
		(*g_pCallback)(IGNITIONEVT, value);
		break;
	
	case CARMSG_Battery_Level_CON:
	case CARMSG_Battery_Level_EVT:
		carapiBlockDecf(f, "v",  &value);
		DBGTRACE(("OnReceiveMessage - BatteryLevel is %d", value));
		(*g_pCallback)(BATCAPACITY, value);
		break;

	case CARMSG_IOKLINE_QueryVersion_CON:{
		FWVERSION v;
		
		carapiBlockDecf(f, "vvvv",  &v.dwVersionDevice, &v.dwVersionFramework, &v.dwVersionCompatibility, &v.dwVersionModule);
		DBGTRACE(("OnReceiveMessage - KLine QueryVersion Device: %d", v.dwVersionDevice));
		DBGTRACE(("OnReceiveMessage - KLine QueryVersion Framework: %d", v.dwVersionFramework));
		DBGTRACE(("OnReceiveMessage - KLine QueryVersion Compatibility: %d", v.dwVersionCompatibility));
		DBGTRACE(("OnReceiveMessage - KLine QueryVersion Module: %d", v.dwVersionModule));
		(*g_pCallback)(VERSION, (LPARAM)&v);
		}
		break;

	case CARMSG_IOKLINE_QueryCaps_CON:{
		BYTE	*pCapsList			= NULL;
		DWORD	capsListLen			= 0;
		BYTE	*pCapsListElem		= NULL;
		DWORD	capsListElemLen		= 0;
		BYTE	*pPhysProps			= NULL;
		DWORD	physPropsLen		= 0;
		BYTE	*pPhysPropsElem		= NULL;
		BYTE	*pLogCapsList		= NULL;
		DWORD	logCapsListLen		= 0;
		BYTE	*pLogCapsListElem	= NULL;
		DWORD	logCapsListElemLen	= 0;

		DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON start decoding."));
		
		//Get pointer to capability-Description list
		DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON carapiBlockDecBlock start"));
		carapiBlockDecBlock(f, &pCapsList, NULL, NULL);
		DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON carapiBlockDecBlock end"));
		assert(pCapsList != NULL);

		//Get pointer to first element in PhysConnCaps-List and lenght of list
		if ( carapiBlockDecBlock(pCapsList, &pCapsListElem, &capsListLen, NULL) ){
			
			assert(pCapsListElem != NULL);
			assert(capsListLen != 0);

			g_syncLock.Aquire();
			
			while( capsListLen > 0 ){

				//Get length of current element in LogConnCaps-List
				carapiBlockGetLength(pCapsListElem, &capsListElemLen);
				assert(capsListElemLen != 0);

				
				carapiBlockDecf(pCapsListElem, "vss",  &g_pm_KLineInfo[i]->physConnectionType,
													   &pPhysProps,
													   &pLogCapsList);
				assert(pPhysProps != NULL);
				assert(pLogCapsList != NULL);

				DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON - physConnType: %d", g_pm_KLineInfo[i]->physConnectionType));

				if ( g_pm_KLineInfo[i]->physConnectionType == 1 ){
					//dont decode PhysConnPropertyList it is empty
					//this may change!!!
				}
				
				//Get pointer to first element in LogConnCaps-List and lenght of list
				carapiBlockDecBlock(pLogCapsList, &pLogCapsListElem, &logCapsListLen, NULL);
				assert(pLogCapsListElem != NULL);

				while( logCapsListLen > 0 ){
					
					//Get length of current element in LogConnCaps-List
					carapiBlockGetLength(pLogCapsListElem, &logCapsListElemLen);
					assert(capsListElemLen != 0);
					
					carapiBlockDecf(pLogCapsListElem, 
									"vvvv", 
									&g_pm_KLineInfo[i]->logConnectionType,
									&g_pm_KLineInfo[i]->busSpeeds,
									&value, //LogConnFlags currently not used
									&g_pm_KLineInfo[i]->txFlags);

					DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON - logConnType: %d", g_pm_KLineInfo[i]->logConnectionType));
					DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON - busSpeeds: %d", g_pm_KLineInfo[i]->busSpeeds));
					DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON - LogFlags: %d", value));
					DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON - txFlags: %d", g_pm_KLineInfo[i]->txFlags));
					
					//set pointer to next element
					pLogCapsListElem += logCapsListElemLen;
					logCapsListLen -= logCapsListElemLen;
				}// end while(logCapsListLen > 0)
				
				//set pointer to next element
				pCapsListElem += capsListElemLen;
				capsListLen -= capsListElemLen;
			}// end while(capsListLen > 0)
			
			g_syncLock.Release();
		}
		else{
			DBGTRACE(("Callback carapiBlockDecBlock failed on pCapsList"));
		}

		DBGTRACE(("Callback CARMSG_IOKLINE_QueryCaps_CON stop decoding."));
		(*g_pCallback)(KCAPSCON, 0);

		}
		break;

	case CARMSG_IOKLINE_PhysicalConnect_CON:

		carapiBlockDecf(f, "v",  &value);
		DBGTRACE(("OnReceiveMessage - KLine-PhysicalConnect_CON for id %d", value));
	
		break;

	case CARMSG_IOKLINE_PhysicalConnect_EVT:
		g_syncLock.Aquire();
		carapiBlockDecf(f, "v",  &g_pm_KLineInfo[i]->physicalId);
		
		//a structured block normally is also shipped with this msg. Currently this
		//structured block is empty so we dont care. In later versions this may change!!!
		
		DBGTRACE(("OnReceiveMessage - KLine-PhysicalConnect_EVT for id %d", g_pm_KLineInfo[i]->physicalId));
		
		g_pm_KLineInfo[i]->connectState = PHYSICAL_RDY;
		g_syncLock.Release();
		
		(*g_pCallback)(KPHYSCONNECT, 0);
		break;
	
	case CARMSG_IOKLINE_LogicalConnect_CON:
		g_syncLock.Aquire();
		carapiBlockDecf(f, "vv", &g_pm_KLineInfo[i]->physicalId, &g_pm_KLineInfo[i]->logicalId);
		g_syncLock.Release();
		DBGTRACE(("OnReceiveMessage - KLine-LogicalConnect_CON for physicalId %d", g_pm_KLineInfo[i]->physicalId));
		DBGTRACE(("OnReceiveMessage - KLine-LogicalConnect_CON for logicalId %d", g_pm_KLineInfo[i]->logicalId));
		break;

	case CARMSG_IOKLINE_LogicalConnect_EVT:
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalConnect_EVT"));
		
		g_syncLock.Aquire();
		carapiBlockDecf(f, 
						"vvvv", 
						&g_pm_KLineInfo[i]->physicalId, 
						&g_pm_KLineInfo[i]->logicalId,
						&g_pm_KLineInfo[i]->logConnectionType,		//unused
						&g_pm_KLineInfo[i]->logConnectionParams);	//unused

		g_pm_KLineInfo[i]->connectState = LOGICAL_RDY;
		g_syncLock.Release();
		
		DBGTRACE(("OnReceiveMessage - KLine-LogicalConnect_EVT for physicalId %d", g_pm_KLineInfo[i]->physicalId));
		DBGTRACE(("OnReceiveMessage - KLine-LogicalConnect_EVT for logicalId %d", g_pm_KLineInfo[i]->logicalId));
		DBGTRACE(("OnReceiveMessage - KLine-LogicalConnect_EVT for logConnectionType %d", g_pm_KLineInfo[i]->logConnectionType));
		DBGTRACE(("OnReceiveMessage - KLine-LogicalConnect_EVT for logConnectionParams %d", g_pm_KLineInfo[i]->logConnectionParams));
		
		(*g_pCallback)(KLOGCONNECT, b);
		break;

	case CARMSG_IOKLINE_LogicalStatus_CON:
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalStatus_CON"));
		break;

	case CARMSG_IOKLINE_LogicalStatus_EVT:
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalStatus_EVT"));
		break;

	case CARMSG_IOKLINE_LogicalDisconnect_CON:
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalDisconnect_CON"));
		
		carapiBlockDecf(f, "vv", &g_pm_KLineInfo[i]->physicalId, &g_pm_KLineInfo[i]->logicalId);

		DBGTRACE(("OnReceiveMessage - KLine-LogicalDisconnect_CON for physicalId %d", g_pm_KLineInfo[i]->physicalId));
		DBGTRACE(("OnReceiveMessage - KLine-LogicalDisconnect_CON for logicalId %d", g_pm_KLineInfo[i]->logicalId));
		break;

	case CARMSG_IOKLINE_LogicalDisconnect_EVT:{
		DWORD physID;
		DWORD logID;
		KLINEANSWER answer;

		carapiBlockDecf(f, "vvv", &physID, &logID, &answer.dwValue);
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalDisconnect_EVT - physID: %d", physID));
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalDisconnect_EVT - logID: %d", logID));
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_LogicalDisconnect_EVT - reason: %d",answer.dwValue));

		answer.uiChannel = b;  //identifies the K-Line-Module that sent the message

		g_syncLock.Aquire();
		g_pm_KLineInfo[i]->disconnectReason = answer.dwValue;
		g_pm_KLineInfo[i]->connectState = PHYSICAL_RDY;
		g_syncLock.Release();
		(*g_pCallback)(KLOGDISCON, (LPARAM)&answer);
		}
		break;
	
	case CARMSG_IOKLINE_PhysicalDisconnect_CON:
		carapiBlockDecf(f, "v", &value);
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_PhysicalDisconnect_CON for id %d", value));
		break;

	case CARMSG_IOKLINE_PhysicalDisconnect_EVT:
		carapiBlockDecf(f, "vv", &g_pm_KLineInfo[i]->physicalId, &value);
		//third value not decoded because currently unused (may change!!!)
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_PhysicalDisconnect_EVT - physID: %d", g_pm_KLineInfo[i]->physicalId));
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_PhysicalDisconnect_EVT - reason: %d", value));
		
		g_syncLock.Aquire();
		g_pm_KLineInfo[i]->connectState = DISCONNECTED;
		g_syncLock.Release();
		(*g_pCallback)(KPHYSDISCON, b);
		break;

	case CARMSG_IOKLINE_Send_CON:
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Send_CON"));
		break;

	case CARMSG_IOKLINE_Recv_EVT:{
		DWORD physId;
		DWORD logId;
		BYTE	*pDataList			= NULL;
		DWORD	dataListLen			= 0;
		BYTE*	pDataListElem		= NULL;
		DWORD   dataListElemLen		= 0;
		BYTE*	pData				= NULL;
		DWORD	dataLen				= 0;
		DWORD	title;
		DWORD   recvFlags;
		DWORD   timeStamp;

		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT"));
		
		carapiBlockDecf(f, "vvs",  &physId, &logId, &pDataList);
		
		DBGTRACE(("OnReceiveMessage - physId: %d", physId));
		DBGTRACE(("OnReceiveMessage - logId: %d", logId));
		
		assert(pDataList != NULL);

		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: physId: %d", physId));
		DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: logId: %d", logId));
		
		//Get pointer to first element in data-List and lenght of list
		carapiBlockDecBlock(pDataList, &pDataListElem, &dataListLen, NULL);
		assert(pDataListElem != NULL);

		while( dataListLen > 0 ){
			
			//Get length of current element in LogConnCaps-List
			carapiBlockGetLength(pDataListElem, &dataListElemLen);
			assert(dataListElemLen != 0);
			
				//decode CAN-Identifier, flags and CAN-message
			carapiBlockDecf(pDataListElem, "vvuv", &recvFlags,
												   &title,
												   &pData, 
												   &dataLen, 
												   &timeStamp);

			DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: title: 0x%x", title));
			DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: recvFlags: 0x%x", recvFlags));
			DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: timeStamp: %d", timeStamp));
			DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: dataLen: 0x%x", dataLen));

			for (unsigned int i = 0; i < dataLen; i++){
				DBGTRACE(("OnReceiveMessage - CARMSG_IOKLINE_Recv_EVT: data: 0x%x", *(pData+i)));
			}

			//set pointer to next element
			pDataListElem += dataListElemLen;
			dataListLen  -= dataListElemLen;
		}// end while
		
		KLINEDATA msg;
		msg.title = (unsigned char)title;
		msg.pData = pData;
		msg.msgLen = dataLen;

		(*g_pCallback)(KRXDATA, (LPARAM)&msg);

		}//end case
		break;

	default:
		DBGTRACE(("Exited from Callback. Couldnt resolve MessageId."));
		break;
	}
	return;
}


/******************************************************************************/
/*
/*  Funktion:		Audio_RequestCaps
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		This function has to be called prior to setting any
/*					sound parameter like volume, fader, bass, treble...
/*					It queries the valid range for all these features.
/*					After the hardware answered with the specific
/*					messages sound parameters can be set.
/*					Sound parameters are hardware dependent and subject to 
/*					change !!!		
/*		
/******************************************************************************/
bool Communication::Audio_RequestCaps(void)
{
	DWORD		  status		= 0;
	unsigned char Message[100]	= {0};

	DBGTRACE(("Audio_Init entered."));

	if ( !m_connected ){
		DBGTRACE(("Audio_Init failed. Entry condition false."));
		return false;
	}

	//build a message that queries the audio sources available
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioSelector_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf selector-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init Encf selector-caps succeeded."));

	//send message
	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send selector-caps."));
		return false;
	}
	
	DBGTRACE(("Audio_Init succeeded request selector-caps."));

	//build a message that queries the volume range
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioVolume_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf volume-caps."));
		return false;
	}
	
	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send volume-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init succeeded request volume-caps."));

	//build a message that queries the balance range
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))",
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioBalance_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf balance-caps."));
		return false;
	}
	
	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send balance-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init succeeded request balance-caps."));

	//build a message that queries the fader range
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))",
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFader_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf fader-caps."));
		return false;
	}
	
	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send fader-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init succeeded request fader-caps."));

	//build a message that queries the bass range
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))",
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFilterBass_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf Bass-caps."));
		return false;
	}
	
	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send Bass-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init succeeded request Bass-caps."));

	//build a message that queries the treble range
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))",
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFilterTreble_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf Treble-caps."));
		return false;
	}
	
	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send Treble-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init succeeded request Treble-caps."));

	//build a message that queries the loudness range
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))",
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFilterLoudness_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Audio_Init failed to Encf Loudness-caps."));
		return false;
	}

	if ( !carapiMsgSend(Message) ){
		DBGTRACE(("Audio_Init failed Send Loudness-caps."));
		return false;
	}

	DBGTRACE(("Audio_Init requested all caps successfully."));
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		Audio_ViewCaps
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:				
/*		
/******************************************************************************/
void Communication::Audio_ViewCaps(AUDIOCAPS /*out*/ *caps)
{
	assert(caps != NULL);
	g_syncLock.Aquire();
	memcpy(caps, &m_AudioCaps, sizeof(m_AudioCaps));
	g_syncLock.Release();
}

/******************************************************************************/
/*
/*  Funktion:		Audio_SetVolume
/*	
/*
/*	Parameters:		value contains the new volume to be set
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after audio capabilities were queried		
/*		
/******************************************************************************/
bool Communication::Audio_SetVolume(int value)
{	
	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Audio_SetVolume entered, request is %d", value));
	
	if ( (!m_connected) || ( m_AudioCaps.volMax == 0) ){
		DBGTRACE(("Audio_SetVolume failed. Entry condition false."));
		return false;
	}

	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioVolume_Position_CMD,
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SetVolume failed to blockEncf"));
		return false;
	}

	DBGTRACE(("Leaving Audio_SetVolume successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Audio_SetBass
/*	
/*
/*	Parameters:		value contains the new bass to be set
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after audio capabilities were queried		
/*		
/******************************************************************************/
bool Communication::Audio_SetBass(int value)
{
	assert( m_AudioCaps.basMax != 0 );

	DWORD		  status		= 0;
	unsigned char Message[64]	= {0};

	DBGTRACE(("Audio_SetBass entered, request is %d", value));
	
	if ( (!m_connected) || ( m_AudioCaps.basMax == 0 ) ){
		DBGTRACE(("Audio_SetBass failed. Entry condition false."));
		return false;
	}
	
	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFilterBass_Position_CON,
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SetBass failed to blockEncf"));
		return false;
	}

	DBGTRACE(("Leaving Audio_SetBass successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Audio_SetTreble
/*	
/*
/*	Parameters:		value contains the new treble to be set
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after audio capabilities were queried		
/*		
/******************************************************************************/
bool Communication::Audio_SetTreble(int value)
{	
	assert( m_AudioCaps.trebMax != 0 );
	
	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Audio_SetTreble entered, request is %d", value));
	
	if ( (!m_connected) || ( m_AudioCaps.trebMax == 0 ) ){
		DBGTRACE(("Audio_SetTreble failed. Entry condition false."));
		return false;
	}

	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFilterTreble_Position_CON,
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SetTreble failed to blockEncf"));
		return false;
	}

	DBGTRACE(("Leaving Audio_SetTreble successful trying to send."));

	return ( SendWithPrio(Message) );
}



/******************************************************************************/
/*
/*  Funktion:		Audio_SetLoudness
/*	
/*
/*	Parameters:		value contains the new loudness to be set
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after audio capabilities were queried	
/*		
/******************************************************************************/
bool Communication::Audio_SetLoudness(int value)
{
	assert( m_AudioCaps.loudMax != 0 );
	
	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Audio_SetLoudness entered, request is %d", value));
	
	if ( (!m_connected) || ( m_AudioCaps.loudMax == 0 ) ){
		DBGTRACE(("Audio_SetLoudness failed. Entry condition false."));
		return false;
	}

	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFilterLoudness_Position_CON,
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SetLoudness failed to blockEncf"));
		return false;
	}

	DBGTRACE(("Leaving Audio_SetLoudness successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Audio_SetBalance
/*	
/*
/*	Parameters:		value contains the new balance to be set
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after audio capabilities were queried	
/*		
/******************************************************************************/
bool Communication::Audio_SetBalance(int value)
{
	assert( m_AudioCaps.balRight != 0 );

	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Audio_SetBalance entered, request is %d", value));
	
	if ( (!m_connected) || ( m_AudioCaps.balRight == 0 ) ){
		DBGTRACE(("Audio_SetBalance failed. Entry condition false."));
		return false;
	}

	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioBalance_Position_CMD,
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SetBalance failed to blockEncf"));
		return false;
	}

	DBGTRACE(("Leaving Audio_SetBalance successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Audio_SetFader
/*	
/*
/*	Parameters:		value contains the new treble to be set
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after audio capabilities were queried	
/*		
/******************************************************************************/
bool Communication::Audio_SetFader(int value)
{	
	assert( m_AudioCaps.fadFront != 0 );

	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Audio_SetFader entered, request is %d", value));
	
	if ( (!m_connected) || ( m_AudioCaps.fadFront == 0 ) ){
		DBGTRACE(("Audio_SetFader failed. Entry condition false."));
		return false;
	}

	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioFader_Position_CMD,
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SetFader failed to blockEncf"));
		return false;
	}

	DBGTRACE(("Leaving Audio_SetFader successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Audio_Mute
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		not used			
/*		
/******************************************************************************/
void Communication::Audio_Mute(bool value)
{
	return;
}


/******************************************************************************/
/*
/*  Funktion:		Audio_SelectSource
/*	
/*
/*	Parameters:		CD				1	
/*					Soundcard (MP3)	2
/*					Phone			3
/*					MM				4
/*					FM Mono			5
/*					FM Stereo		6
/*					Mute			7
/*
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after source capabilities were 
/*					queried
/*								
/******************************************************************************/
bool Communication::Audio_SelectSource(int value)
{	
	assert(value >= 0);
	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};
	
	DBGTRACE(("Audio_SelectSource entered requesting Src %d.", value));

	if ( (!m_connected) ){
		DBGTRACE(("Audio_SelectSource failed. Entry condition false."));
		return false;
	}

	//encode message
	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv(v)))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioSelector_Position_CMD, 
							0, 
							value);

	if ( 0 == status ){
		DBGTRACE(("Audio_SelectSource Encf failed."));
		return false;
	}

	DBGTRACE(("Leaving Audio_SelectSource successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_RequestCaps
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Sending a CARMSG_AudioTuner_QueryCaps_CMD is mandatory be-
/*					fore other tuner-related messages are accepted. Otherwise
/*					these messages will return with an error!		
/*		
/******************************************************************************/
bool Communication::Tuner_RequestCaps(void)
{	
	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Tuner_RequestCaps entered."));

	if ( (!m_connected) ){
		DBGTRACE(("Tuner_RequestCaps failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(Message, 
							sizeof(Message), 
							"v((vvvv()))", 
							m_linkId, 
							AUDIOMODUL, 
							0, 
							CARMSG_AudioTuner_QueryCaps_CMD,
							0);
	
	if ( 0 == status ){
		DBGTRACE(("Tuner_RequestCaps Encf failed."));
		return false;
	}

	DBGTRACE(("Leaving Tuner_RequestCaps successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_ViewCaps
/*	
/*
/*	Parameters:		AreaDescription	  pointer to area description struct
/*					Area			  id of area description
/*
/*	Return Value:	--
/*
/*	Comments:				
/*		
/******************************************************************************/
void Communication::Tuner_ViewCaps(/*out*/ AreaDescription *caps, int area)
{	
	assert(caps != NULL);

	//valid areas: 1, 2, 3, 4
	assert( (area > 0) && (area < 5) );

	//Areas are stored in an array with zero based index count
	area--;

	g_syncLock.Aquire();
	*caps = g_areaDescr[area];
	g_syncLock.Release();
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_GetFrequencyParams
/*	
/*
/*	Parameters:		--
/*					
/*
/*	Return Value:	--
/*
/*	Comments:		--		
/*		
/******************************************************************************/
void Communication::Tuner_GetFrequencyParams(/*out*/ FREQPARAMS *params)
{	
	assert(params != NULL);
	g_syncLock.Aquire();
	memcpy(params, &g_freqParams, sizeof(g_freqParams));
	g_syncLock.Release();
}

/******************************************************************************/
/*
/*  Funktion:		Tuner_SetFrequency
/*	
/*
/*	Parameters:		value	frequency in kHz (102,6 MHz -> 10260)
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The accepted frequency range depends on the current setting
/*					of area (EUROPE, USA...) and the band (FM, SW...).
/*					If not accepted CARMSG_AudioTuner_SetFrequency_CON returns
/*					with an error.
/*		
/******************************************************************************/
bool Communication::Tuner_SetFrequency(int value)
{	
	DBGTRACE(("Tuner_SetFrequency entered, freq request is %d", value));

	if ( (!m_connected) ){	
		DBGTRACE(("Tuner_SetFrequency failed. Entry condition false."));
		return false;
	}

	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};
	
	DBGTRACE(("Tuner_SetFrequency, requested f is %d", value));

	status = carapiBlockEncf(Message, 
						sizeof(Message), 
						"v((vvvv(v)))", 
						m_linkId, 
						AUDIOMODUL, 
						0, 
						CARMSG_AudioTuner_SetFrequency_CMD,
						0, 
						value);

	if ( 0 == status ){
		return false;
	}

	DBGTRACE(("Leaving Tuner_SetFrquency successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_SelectBand
/*	
/*
/*	Parameters:		band to be set
/*					1	FM
/*					2	SW
/*					3	MW
/*					4	LW
/*					5	WX
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The message sent here will only be accepted if a
/*					CARMSG_AudioTuner_QueryCaps_CMD was sent before which also
/*					returns information about the supported bands.
/*					Supported bands are subject to change!!!
/*		
/******************************************************************************/
bool Communication::Tuner_SelectBand(int value)
{
	DBGTRACE(("Tuner_SelectBand entered, band request is %d", value));

	if ( !m_connected )
	{	
		DBGTRACE(("Tuner_SelectBand failed. Entry condition false."));
		return false;
	}

	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Tuner_SelectBand, requested band is %d", value));

	status = carapiBlockEncf(Message, 
						sizeof(Message), 
						"v((vvvv(v)))", 
						m_linkId, 
						AUDIOMODUL, 
						0, 
						CARMSG_AudioTuner_SelectBand_CMD,
						0, 
						value);

	if ( 0 == status ){
		return false;
	}

	DBGTRACE(("Leaving Tuner_SelectBand successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_ScanBand
/*	
/*
/*	Parameters:		minFreq		start of scan in kHz
/*					maxFreq		end of scan in kHz
/*					step		distance between scan samples in kHz
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		acceptance of message-parameters depend on current
/*					setting	of area (EUROPE, USA...) and the band (FM, SW...).		
/*		
/******************************************************************************/
bool Communication::Tuner_ScanBand(int minFreq, int maxFreq, int step)
{	
	DBGTRACE(("Tuner_ScanBand entered, min:%d  max:%d  step:%d", 
		minFreq, maxFreq, step));

	if ( !m_connected ){	
		DBGTRACE(("Tuner_ScanBand failed. Entry condition false."));
		return false;
	}

	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	status = carapiBlockEncf(Message, 
						sizeof(Message), 
						"v((vvvv(vvv)))", 
						m_linkId, 
						AUDIOMODUL, 
						0, 
						CARMSG_AudioTuner_ScanBand_CMD,
						0, 
						minFreq,
						maxFreq,
						step);

	if ( 0 == status ){
		return false;
	}

	DBGTRACE(("Leaving Tuner_ScanBand successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_SelectArea
/*	
/*
/*	Parameters:		area to be set
/*					1	EUROPE
/*					2	USA
/*					3	EAST
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Does only work after a CARMSG_AudioTuner_QueryCaps_CMD was
/*					sent. 		
/*		
/******************************************************************************/
bool Communication::Tuner_SelectArea(int value)
{	
	DBGTRACE(("Tuner_SelectArea entered, area request is %d", value));

	if ( !m_connected ){
		DBGTRACE(("Tuner_SelectArea failed. Entry condition false."));
		return false;
	}

	DWORD		  status	  = 0;
	unsigned char Message[64] = {0};

	DBGTRACE(("Tuner_SelectArea, requested area is %d", value));

	status = carapiBlockEncf(Message, 
						sizeof(Message), 
						"v((vvvv(v)))", 
						m_linkId, 
						AUDIOMODUL,
						0,
						CARMSG_AudioTuner_SelectArea_CMD,
						0, 
						value);

	if ( 0 == status ){
		return false;
	}

	DBGTRACE(("Leaving Tuner_SelectArea successful trying to send."));

	return ( SendWithPrio(Message) );
}


/******************************************************************************/
/*
/*  Funktion:		Tuner_GetScanResult
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		--			
/*		
/******************************************************************************/
void Communication::Tuner_GetScanResult(FREQULIST *list)
{
	assert( list != NULL);
	memcpy(list, &g_scanList, sizeof(g_scanList) );
}

/******************************************************************************/
/*
/*  Funktion:		CAN_PhysicalConnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		The confirmation on the request for a physical connect will 
/*					return with an error if the physical connection is already
/*					established!				
/*		
/******************************************************************************/
bool Communication::CAN_PhysicalConnect(int physConId, 
										int conType, 
										int busSpeed)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	if ( !m_connected ){
		DBGTRACE(("CAN_PhysicalConnect - failed. Entry condition false."));
		return false;
	}

	if ( m_CANInfo.connectState != DISCONNECTED) {
		DBGTRACE(("CAN_LogicalConnect - failed. CAN not disconnected."));
		return false;
	}

	DBGTRACE(("CAN_PhysicalConnect - entered"));
	DBGTRACE(("CAN_PhysicalDisconnect - try to mount physId %d.", physConId));
	DBGTRACE(("CAN_PhysicalConnect - connectType: 0x%x   baud: %d", conType, busSpeed));
	
	status = carapiBlockEncf(msgBuf,
							 sizeof(msgBuf),
							 "v((vvvv(vv(v))))", 
							 m_linkId, 
							 CANMODUL,
							 0, 
							 CARMSG_IOCAN_PhysicalConnect_CMD,
							 0,
							 physConId,
							 conType, 
							 busSpeed
							 );

	if ( 0 == status ) 
	{
		DBGTRACE(("CAN_PhysicalConnect - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("CAN_PhysicalConnect - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("CAN_PhysicalConnect - msgSend successful"));	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		CAN_LogicalConnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		The confirmation on the request for a logical connect will 
/*					return with an error if the logical connection is already
/*					established! 				
/*		
/******************************************************************************/
DWORD Communication::CAN_LogicalConnect(DWORD logicalConId,
										 DWORD acceptanceMask, 
										 DWORD acceptanceCode,
										 BYTE  connectionflags,
										 BYTE  mailboxFlags)
{
	DBGTRACE(("CAN_LogicalConnect - entered"));

	if ( !m_connected ){
		DBGTRACE(("CAN_LogicalConnect failed. Entry condition false."));
		return false;
	}
	
	if ( m_CANInfo.connectState != PHYSICAL_RDY) {
		DBGTRACE(("CAN_LogicalConnect failed. No physical connection."));
		return false;
	}

	DWORD			status		= 0;
	unsigned char	msgBuf[128]	= {0};
	
		
	DBGTRACE(("CAN_PhysicalDisconnect - try to mount logId %d in physId %d.", 
				logicalConId,
				m_CANInfo.physicalId));
	DBGTRACE(("CAN_LogicalConnect - acceptanceMask: 0x%08x", acceptanceMask));
	DBGTRACE(("CAN_LogicalConnect - acceptanceCode: 0x%08x", acceptanceCode));
	DBGTRACE(("CAN_LogicalConnect - connectionflags: 0x%02x", connectionflags));
	DBGTRACE(("CAN_LogicalConnect - mailboxFlags: 0x%02x", mailboxFlags));

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv(vv(vv((vvv(vv)))))))",
							 m_linkId, 
							 CANMODUL,		
							 0,		
							 CARMSG_IOCAN_LogicalConnect_CMD,
							 0,				
							 m_CANInfo.physicalId,		// PhysConID
							 logicalConId,		// LogConnID
							 connectionflags,	// flags 1
							 5,					// timeout
							 mailboxFlags,		// flags 2
							 acceptanceMask,	// acceptance mask
							 acceptanceCode,	// acceptance code
							 0xdf,				// DataByte1 mask
							 0xab				// DataByte1 code
							 );
/*
	//definiert zwei Filter (Test)
	//Filter 1: id 0xaa wird ausgefiltert
	//Filter 2: Bereich von id 0x374 - 0x375 wird ausgefiltert
	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv(vv(vv((vvvvv)(vvvvv))))))",
							 m_linkId, 
							 g_deviceIdCAN,		// CAN id
							 0,		// message handle
							 CARMSG_IOCAN_LogicalConnect_CMD,
							 0,		// status
							 1,		// PhysConID
							 logicalConId,		// LogConnID
							 connectionflags,	// flags 1
							 5,		// timeout
							 0x01,	//ignorieren
							 0x1fffff00,	// acceptance mask
							 0xaa,	// acceptance code
							 0xdf,		// DataByte1 mask
							 0xab,		// DataByte1 code
							 0x01,
							 0x1ffff801,
							 0x394,
							 0xc3,
							 0xc4
							 );
*/	


	if( 0 == status ){
		DBGTRACE(("CAN_LogicalConnect - encf failed"));	
		return status;
	}
		
	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("CAN_LogicalConnect - msgSend failed"));	
		return status;
	}
	
	DBGTRACE(("CAN_LogicalConnect - msgSend successful"));
	return status;
}


/*****************************************************************************/
/*
/*  Funktion:		CAN_SendMessage
/*	
/*
/*	Parameters:		CANId		identifier of CAN message
/*					ack			not used
/*					rtr			indicates a remote transmission
/*					conf		requests a CARMSG_IOCAN_Send_CON message
/*					pCANMsg		pointer to CAN message bytes
/*					CANMsgLen	length of CAN message
/*
/*	Return Value:	--
/*
/*	Comments:				
/*		
/*****************************************************************************/
bool Communication::CAN_SendMessage(DWORD			CANId,
									 BOOL			ack,
									 BOOL			rtr,
									 BOOL			conf,
									 unsigned char *pCANMsg, 
									 int			CANMsgLen)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[128]	= {0};
	DWORD			msgFlags	= 0;
	
	DBGTRACE(("CAN_SendMessage - entered"));

	if ( !m_connected ){
		DBGTRACE(("CAN_SendMessage - failed. Entry condition false."));
		return false;
	}
	
	if ( m_CANInfo.connectState != LOGICAL_RDY) {
		DBGTRACE(("CAN_SendMessage - failed. No logical connection."));
		return false;
	}

	if ( pCANMsg == NULL ){
		DBGTRACE(("CAN_SendMessage - failed. CAN message NULL."));
		return false;
	}

	if ( CANMsgLen > 8 ){
		DBGTRACE(("CAN_SendMessage - failed. invalid CAN message length."));
		return false;
	}

	ack <<= 1;
	conf <<= 3;

	msgFlags = (ack | rtr | conf);
	DBGTRACE(("CAN_SendMessage - msgFlags: 0x%x", msgFlags));
	DBGTRACE(("CAN_SendMessage - CAN ACK-Bit: 0x%x", ack));
	DBGTRACE(("CAN_SendMessage - CAN RTR-Bit: 0x%x", rtr));
	DBGTRACE(("CAN_SendMessage - CAN CONF-Bit: 0x%x", conf));
	DBGTRACE(("CAN_SendMessage - CAN Id: 0x%x", CANId));
	DBGTRACE(("CAN_SendMessage - CAN Length: 0x%x", CANMsgLen));

#ifdef _DEBUG
			TCHAR stream[32] = {0};

			for (int i = 0; i < CANMsgLen; i++){
				sprintf(stream, "%02x ", *(pCANMsg+i));
			}
			DBGTRACE(("CAN_SendMessage - CAN Msg: %s", stream));
#endif

	status = carapiBlockEncf(msgBuf,
							sizeof(msgBuf),
							"v((vvvv(vv((vvu)))))", 
							m_linkId, 
							CANMODUL,
							0, 
							CARMSG_IOCAN_Send_CMD,
							0,
							m_CANInfo.physicalId,	// PhysCon ID
							m_CANInfo.logicalId,	// logicalCon ID
							CANId,					// CAN message ID
							msgFlags,	
							pCANMsg,
							CANMsgLen
							);

	if ( 0 == status ){
		DBGTRACE(("CAN_SendMessage - encf failed"));	
		return false;
	}
	else{
		DBGTRACE(("CAN_SendMessage - encf success"));
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("CAN_SendMessage - msgSend failed"));	
		return false;
	}

	DBGTRACE(("CAN_SendMessage - msgSend success"));
	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		CAN_PhysicalDisconnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:				
/*		
/******************************************************************************/
bool Communication::CAN_PhysicalDisconnect(void)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("CAN_PhysicalDisconnect - entered"));

	if ( !m_connected ){
		DBGTRACE(("CAN_PhysicalDisconnect - failed. Entry condition false."));
		return false;
	}

	if ( m_CANInfo.connectState != PHYSICAL_RDY) {
		DBGTRACE(("CAN_PhysicalDisconnect - failed. No logical connection."));
		return false;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv(v)))", 
							 m_linkId, 
							 CANMODUL,	// CAN id
							 0,			// message handle
							 CARMSG_IOCAN_PhysicalDisconnect_CMD,
							 0,			// status
							 m_CANInfo.physicalId // PhysConID
							 );

	if( 0 == status ){
		DBGTRACE(("CAN_PhysicalDisconnect - encf failed"));	
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("CAN_PhysicalDisconnect - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("CAN_PhysicalDisconnect - msgSend successful"));
	return true;
}
	

/******************************************************************************/
/*
/*  Funktion:		CAN_LogicalDisconnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		--		
/*		
/******************************************************************************/
DWORD Communication::CAN_LogicalDisconnect(void)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};

	DBGTRACE(("CAN_LogicalDisconnect - entered"));
	
	if ( !m_connected ){
		DBGTRACE(("CAN_LogicalDisconnect - failed. Entry condition false."));
		return status;
	}

	if ( m_CANInfo.connectState != LOGICAL_RDY) {
		DBGTRACE(("CAN_LogicalDisconnect - failed. No logical connection."));
		return status;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv(vv)))", 
							 m_linkId, 
							 CANMODUL,	// CAN id
							 0,			// message handle
							 CARMSG_IOCAN_LogicalDisconnect_CMD,
							 0,			// status
							 m_CANInfo.physicalId,	// PhysConID
							 m_CANInfo.logicalId	// LogConnID
							 );

	if( 0 == status ){
		DBGTRACE(("CAN_LogicalDisconnect - encf failed"));	
		return status;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("CAN_LogicalDisconnect - msgSend failed"));	
		return status;
	}
	
	DBGTRACE(("CAN_LogicalDisconnect - msgSend successful"));
	return status;
}


/******************************************************************************/
/*
/*  Funktion:		CAN_ViewConnectionInfo
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:				
/*		
/******************************************************************************/
void  Communication::CAN_ViewConnectionInfo(CANINFO *pInfo)
{	
	g_syncLock.Aquire();
	memcpy(pInfo, &m_CANInfo, sizeof(CANINFO));
	g_syncLock.Release();
}


/******************************************************************************/
/*
/*  Funktion:		Power_SwitchOffSupply
/*	
/*
/*	Parameters:		switch off delay in milliseconds
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The system will be switched off after the given 
/*					delay. When sending CARMSG_PowerSupply_SwitchOff_CMD
/*					the operating system has to be shut down within the
/*					given switch-off delay. Else data loss or system 
/*					damage may occur!!! It is not possible to disable
/*					the switch-off sequence once it was triggered. 		
/*		
/******************************************************************************/
bool Communication::Power_SwitchOffSupply(unsigned int delay)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("SwitchOffPowerSupply - entered"));

	if ( !m_connected ){
		DBGTRACE(("Power_SwitchOffSupply - failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv(v)))", 
							 m_linkId, 
							 POWERMODUL,
							 0, 
							 CARMSG_PowerSupply_SwitchOff_CMD,
							 0, 
							 delay // switch off after n milliseconds
							 );

	if( 0 == status ){
		DBGTRACE(("Power_SwitchOffSupply - encf failed"));	
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("Power_SwitchOffSupply - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("Power_SwitchOffSupply - msgSend successful"));
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		Power_QueryTerm15
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		Queries the ignition state		
/*		
/******************************************************************************/
bool Communication::Power_QueryTerm15(void)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("Power_QueryTerm15 - entered"));

	if ( !m_connected ){
		DBGTRACE(("Power_QueryTerm15 - failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv()))", 
							 m_linkId, 
							 POWERMODUL,
							 0, // message handle
							 CARMSG_Ignition_State_CMD,
							 0 // status
							 );

	if( 0 == status ){
		DBGTRACE(("Power_QueryTerm15 - encf failed"));	
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("Power_QueryTerm15 - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("Power_QueryTerm15 - msgSend successful"));
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		Power_QueryCapacity
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		Answer to this message returns remaining capacity in %		
/*		
/******************************************************************************/
bool Communication::Power_QueryCapacity(void)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("Power_QueryCapacity - entered"));

	if ( !m_connected ){
		DBGTRACE(("Power_QueryCapacity - failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv()))", 
							 m_linkId, 
							 POWERMODUL,
							 0, // message handle
							 CARMSG_Battery_Level_CMD,
							 0 // status
							 );

	if( 0 == status ){
		DBGTRACE(("Power_QueryCapacity - encf failed"));	
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("Power_QueryCapacity - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("Power_QueryCapacity - msgSend successful"));
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		QueryCurrentValue
/*	
/*
/*	Parameters:		nItem	query what?
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		-- 			
/*		
/******************************************************************************/
bool Communication::QueryCurrentValue(unsigned int nItem)
{	
	DBGTRACE(("GetActualValue entered, requestet state is %d", nItem));

	if ( !m_connected ){	
		DBGTRACE(("GetActualValue failed. Entry condition false."));
		return false;
	}

	DWORD request = 0;
	
	switch (nItem){
	
	case VOLSTAT:
		request = CARMSG_AudioVolume_Position_CMD;
		break;
		
	case BALSTAT:
		request = CARMSG_AudioBalance_Position_CMD;
		break;
		
	case FADSTAT:
		request = CARMSG_AudioFader_Position_CMD;
		break;

	case BASSTAT:
		request = CARMSG_AudioFilterBass_Position_CMD;
		break;

	case TREBSTAT:
		request = CARMSG_AudioFilterTreble_Position_CMD;
		break;

	case LOUDSTAT:
		request = CARMSG_AudioFilterLoudness_Position_CMD;
		break;

	case FREQSTAT:
		request = CARMSG_AudioTuner_SetFrequency_CMD;
		break;

	case BANDSTAT:
		request = CARMSG_AudioTuner_SelectBand_CMD;
		break;

	case AREASTAT:
		request = CARMSG_AudioTuner_SelectArea_CMD;
		break;

	case SRCSTAT:
		request = CARMSG_AudioSelector_Position_CMD;
		break;

	default:
		assert(false);
		DBGTRACE(("GetActualValue couldnt resolve request."));
		return false;
	}

	//encode message
	unsigned char Message[64] = {0};
	DWORD status = carapiBlockEncf(Message, 
						sizeof(Message), 
						"v((vvvv()))",
						m_linkId, 
						AUDIOMODUL, 
						0, 
						request,
						0);

	if ( 0 == status ){
		DBGTRACE(("GetActualValue BlockEncf failed."));
		return false;
	}

	DBGTRACE(("Leaving GetActualValue successful trying to send."));

	return ( SendWithPrio(Message) );
}

/******************************************************************************/
/*
/*  Funktion:		SendUserDefinedMessage
/*	
/*
/*	Parameters:		pMessage	pointer to message
/*					len			length of message
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		development purpose only!		
/*		
/******************************************************************************/
bool Communication::SendUserDefinedMessage(unsigned char *pMessage, unsigned int len)
{	
	assert( (pMessage != NULL) && (len != 0) );

	DBGTRACE(("SendUserDefinedMessage entered. Size is %d", len));

	if ( !m_connected ){
		DBGTRACE(("SendUserDefinedMessage failed. Entry condition false."));
		return false;
	}

	int size = (len + 64);
	
	unsigned char *Message = new unsigned char[size];

	if ( NULL == Message ){
		DBGTRACE(("SendUserDefinedMessage not enough memory"));
		return false;
	}
	
	//encode message
	unsigned char success = carapiBlockEncf(Message, 
											size, 
											"v((vvvv(u)))", 
											m_linkId, 
											0x20000, 
											0, 
											CARMSG_AudioVolume_Position_CMD,
											0, 
											pMessage,
											len);

	delete [] Message;

	if ( !success ){
		DBGTRACE(("SendUserDefinedMessage Error BlockEncf"));
		return false;
	}
	else{
		DBGTRACE(("SendUserDefinedMessage successful, try sending."));
		success = SendWithPrio(Message);
		return true;
	}

}


/******************************************************************************/
/*
/*  Funktion:		SendPCStartUpMessage
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		This will send a message indicating that the PC has
/*					started and some application is able to handle power
/*					management messages. - Very important!!!
/*		
/******************************************************************************/
bool Communication::SendPCStartUpMessage(void)
{	
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("Power_QueryTerm15 - entered"));

	if ( !m_connected ){
		DBGTRACE(("SendPCStartUpMessage - failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv()))", 
							 m_linkId, 
							 POWERMODUL,
							 0, // message handle
							 CARMSG_RegisterClient_CMD,
							 0 // status
							 );

	if( 0 == status ){
		DBGTRACE(("SendPCStartUpMessage - encf failed"));	
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("SendPCStartUpMessage - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("SendPCStartUpMessage - msgSend successful"));
	return true;
	
}


/******************************************************************************/
/*
/*  Funktion:		KLine_QueryVersion
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		--
/*		
/******************************************************************************/
bool Communication::KLine_QueryVersion(void)
{	
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("KLine_QueryVersion - entered"));

	if ( !m_connected ){
		DBGTRACE(("KLine_QueryVersion - failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(msgBuf, 
							 sizeof(msgBuf), 
							 "v((vvvv()))", 
							 m_linkId, 
							 KLINEMODUL1,
							 0, // message handle
							 CARMSG_IOKLINE_QueryVersion_CMD,
							 0 // status
							 );

	if( 0 == status ){
		DBGTRACE(("KLine_QueryVersion - encf failed"));	
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if ( 0 == status ){
		DBGTRACE(("KLine_QueryVersion - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_QueryVersion - msgSend successful"));
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_ViewConnectionInfo
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		--
/*		
/******************************************************************************/
void Communication::KLine_ViewConnectionInfo(IN	int	nChannel, OUT KLINEINFO_P pInfo)
{	
	assert(pInfo != NULL);

	int i = FindKLineChannel(nChannel);
	
	g_syncLock.Aquire();
	memcpy(pInfo, &m_KLineInfo[i], sizeof(KLINEINFO)); 
	g_syncLock.Release();
	return;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_RequestCapabilities
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The message CARMSG_IOKLINE_QueryCaps_CMD has not to be sent
/*					but it returns information about hardware capabilities
/*					that might be useful when setting up a connection.
/*					Hardware capabilities are subject to change!!! 
/*		
/******************************************************************************/
bool Communication::KLine_RequestCapabilities(int kLineChannel)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	DBGTRACE(("KLine_RequestCapabilities - Channel: %d", kLineChannel));

	if ( !m_connected ){
		DBGTRACE(("KLine_RequestCapabilities - failed. Entry condition false."));
		return false;
	}
	
	DBGTRACE(("KLine_RequestCapabilities - entered"));

	status = carapiBlockEncf(msgBuf,
							 sizeof(msgBuf),
							 "v((vvvv()))", 
							 m_linkId, 
							 kLineChannel,
							 0, 
							 CARMSG_IOKLINE_QueryCaps_CMD
							 );

	if ( 0 == status ) 
	{
		DBGTRACE(("KLine_RequestCapabilities - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("KLine_RequestCapabilities - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_RequestCapabilities - msgSend successful"));	
	return true;

	return false;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_PhysicalConnect
/*	
/*
/*	Parameters:		physID	take a value of 1
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The confirmation on the request for a physical connect will 
/*					return with an error if a physical or logical connection is 
/*					already	established! 
/*		
/******************************************************************************/
bool Communication::KLine_PhysicalConnect(int nChannel, int physId)
{	
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	if ( !m_connected ){
		DBGTRACE(("KLine_PhysicalConnect - failed. Entry condition false."));
		return false;
	}

	DBGTRACE(("KLine_PhysicalConnect - entered"));
	DBGTRACE(("KLine_PhysicalConnect - try to mount physId %d.", physId));
	status = carapiBlockEncf(msgBuf,
							 sizeof(msgBuf),
							 "v((vvvv(vv(v))))", 
							 m_linkId, 
							 nChannel,
							 0, 
							 CARMSG_IOKLINE_PhysicalConnect_CMD,
							 0,
							 physId,
							 1,
							 0
							 );

	if ( 0 == status ) 
	{
		DBGTRACE(("KLine_PhysicalConnect - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("KLine_PhysicalConnect - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_PhysicalConnect - msgSend successful"));	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_PhysicalDisconnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The answer to a physical disconnect request will return with 
/*					an error if no connection or a logical connection is
/*					currently established.
/*		
/******************************************************************************/
bool Communication::KLine_PhysicalDisconnect(IN int nChannel, IN int physConId)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	if ( !m_connected ){
		DBGTRACE(("KLine_PhysicalDisconnect - failed. Entry condition false."));
		return false;
	}

	status = carapiBlockEncf(msgBuf,
							 sizeof(msgBuf),
							 "v((vvvv(vv(v))))", 
							 m_linkId, 
							 nChannel,
							 0, 
							 CARMSG_IOKLINE_PhysicalDisconnect_CMD,
							 0,
							 physConId
							 );

	if ( 0 == status ) 
	{
		DBGTRACE(("KLine_PhysicalDisconnect - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("KLine_PhysicalDisconnect - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_PhysicalDisconnect - msgSend successful"));	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_LogicalConnect
/*	
/*
/*	Parameters:		physID	take value returned by PhysicalConnect_CON
/*					
/*					logID	logical-ID  take 1
/*					
/*					protocolID
/*					1		KWP71
/*					2		KWP2000 (not supported)
/*					
/*					bitRate
/*					0		autodetect (not supported yet)
/*					2400	[Bit/s]
/*					4800	[Bit/s]
/*					9600	[Bit/s]
/*					10400	[Bit/s]
/*
/*					nodeID	ID of K-Line device
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The confirmation on the request for a logical connect will 
/*					return with an error if the logical connection is already
/*					established! 
/*		
/******************************************************************************/
bool Communication::KLine_LogicalConnect(int nChannel,   int physId,  int logId, 
										 int protocolID, int bitRate, int nodeID)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	if ( !m_connected ){
		DBGTRACE(("KLine_LogicalConnect - failed. Entry condition false."));
		return false;
	}
	
	DBGTRACE(("KLine_LogicalConnect - nChannel %d", nChannel));
	DBGTRACE(("KLine_LogicalConnect - physId %d", physId));
	DBGTRACE(("KLine_LogicalConnect - logId %d", logId));
	DBGTRACE(("KLine_LogicalConnect - protocolID %d", protocolID));
	DBGTRACE(("KLine_LogicalConnect - bitRate %d", bitRate));
	DBGTRACE(("KLine_LogicalConnect - nodeID %d", nodeID));
	
	status = carapiBlockEncf(msgBuf,
							 sizeof(msgBuf),
							 "v((vvvv(vvv(vv))))", 
							 m_linkId, 
							 nChannel,
							 0, 
							 CARMSG_IOKLINE_LogicalConnect_CMD,
							 0,
							 physId,
							 logId,
							 protocolID, //must be KWP71 others are not supported currently
							 bitRate,
							 nodeID
							 );

	if ( 0 == status ) 
	{
		DBGTRACE(("KLine_LogicalConnect - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("KLine_LogicalConnect - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_LogicalConnect - msgSend successful"));	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_SendData
/*	
/*
/*	Parameters:		channel
/*					5	K-Line 1
/*					6   K-Line 2
/*
/*					physConID	
/*					take value returned by PhysicalConnect_CON
/*
/*					logConId
/*					take value returned by LogicalConnect_CON
/*
/*					pData
/*					Data to send (does not contain block count or end marker)
/*
/*					blockId
/*					Block identifier
/*
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		Works only if LogicalConnect_CON returned without error
/*		
/******************************************************************************/
bool Communication::KLine_SendData(int channel,  int   physConId, int   logConId, 
								   char blockId, char  flags,     char* pData,	  
								   DWORD dataLen)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[128]	= {0};
	
	if ( !m_connected ){
		DBGTRACE(("KLine_SendData - failed. Entry condition false."));
		return false;
	}
	
	DBGTRACE(("KLine_SendData - channel %d", channel));
	DBGTRACE(("KLine_SendData - physConId %d", physConId));
	DBGTRACE(("KLine_SendData - logConId %d", logConId));
	DBGTRACE(("KLine_SendData - blockId 0x%x", blockId));
	DBGTRACE(("KLine_SendData - flags 0x%x", flags));
	DBGTRACE(("KLine_SendData - dataLen %d", dataLen));
	
		if ( dataLen > 0 ){
			status = carapiBlockEncf(msgBuf,
									 sizeof(msgBuf),
									 "v((vvvv(vv((vvu)))))", 
									 m_linkId, 
									 channel,
									 0, 
									 CARMSG_IOKLINE_Send_CMD,
									 0,
									 physConId,
									 logConId,
									 flags,		
									 blockId,
									 pData,		
									 dataLen
									 );
		}
		else{
				status = carapiBlockEncf(msgBuf,
									 sizeof(msgBuf),
									 "v((vvvv(vv((vv)))))", 
									 m_linkId, 
									 channel,
									 0, 
									 CARMSG_IOKLINE_Send_CMD,
									 0,
									 physConId,
									 logConId,
									 flags,		
									 blockId
									 );
		}
/*

	status = carapiBlockEncf(msgBuf,
							sizeof(msgBuf),
							"v((vvvv(vv((vvu)))))",
							m_linkId, 
							CANMODUL,
							0, 
							CARMSG_IOCAN_Send_CMD,
							0,
							m_CANInfo.physicalId,	// PhysCon ID
							m_CANInfo.logicalId,	// logicalCon ID
							CANId,					// CAN message ID
							msgFlags,	
							pCANMsg,
							CANMsgLen
							);


  */
	if ( 0 == status ) 
	{
		DBGTRACE(("KLine_SendData - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("KLine_SendData - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_SendData - msgSend successful"));	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		KLine_LogicalDisconnect
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		The answer to this message will return with error, if
/*					the logical connetion was not established. 
/*		
/******************************************************************************/
bool Communication::KLine_LogicalDisconnect(IN int nChannel, IN int physConnId,  IN int logConnId)
{
	DWORD			status		= 0;
	unsigned char	msgBuf[64]	= {0};
	
	if ( !m_connected ){
		DBGTRACE(("KLine_LogicalDisconnect - failed. Entry condition false."));
		return false;
	}
	
	status = carapiBlockEncf(msgBuf,
							 sizeof(msgBuf),
							 "v((vvvv(vv)))", 
							 m_linkId, 
							 nChannel,
							 0, 
							 CARMSG_IOKLINE_LogicalDisconnect_CMD,
							 0,
							 physConnId,
							 logConnId
							 );

	if ( 0 == status ) 
	{
		DBGTRACE(("KLine_LogicalDisconnect - encf failed"));
		return false;
	}

	status = carapiMsgSend(msgBuf);

	if( 0 == status ){
		DBGTRACE(("KLine_LogicalDisconnect - msgSend failed"));	
		return false;
	}
	
	DBGTRACE(("KLine_LogicalDisconnect - msgSend successful"));	
	return true;
}


/******************************************************************************/
/*
/*  Funktion:		SendWithPrio
/*	
/*
/*	Parameters:		pMessage	pointer to message
/*
/*	Return Value:	false on error true else
/*
/*	Comments:		--
/*		
/******************************************************************************/
bool Communication::SendWithPrio(unsigned char* pMessage)
{	
	assert(pMessage != NULL);

	//Prio has to be one of 1 2 or 3
	assert( (m_CurrentPrio > 0) && (m_CurrentPrio < 4) );
	
	unsigned char success = false;

	switch (m_CurrentPrio){
	
	case 1:
		DBGTRACE(("Prio is high."));
		success = carapiMsgSendHighPri(pMessage);
		break;

	case 2:
		DBGTRACE(("Prio is medium."));
		success = carapiMsgSend(pMessage);
		break;

	case 3: 
		DBGTRACE(("Prio is low."));
		success = carapiMsgSendLowPri(pMessage);
		break;

	default:
		assert(false);
		DBGTRACE(("SendWithPrio unresolved prio."));
		success = false;
	}

	if ( success ){
		DBGTRACE(("SendWithPrio succeeded."));
		return true;
	}
	else{
		DBGTRACE(("SendWithPrio failed."));
		return false;
	}
}


/******************************************************************************/
/*
/*  Funktion:		FindKLineChannel
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		maps module ID to an array index for K-Line data struct
/*		
/******************************************************************************/
int Communication::FindKLineChannel(int modulID)
{
	assert( (modulID != 0) && (modulID != 1) );

	if ( modulID == KLINEMODUL1 ){
		return 0;
	}
	else{
		return 1;
	}
}


/******************************************************************************/
/*
/*  Funktion:		SetDialTarget
/*	
/*
/*	Parameters:		target: das Ziel fr DrDr-Events
/*					startVal: der erste Wert fr target
/*					multipl: legt fest, wie schnell obere und untere
/*							 Grenze erreicht werden kann
/*
/*	Return Value:	--
/*
/*	Comments:		development purpose only!	
/*		
/******************************************************************************/
void Communication::SetDialTarget(int target, int startVal, int multipl, bool bUseSt)
{			
	assert( target >= 0 );
	DBGTRACE(("SetDialTarget tar %d, start %d, mul %d", target, startVal, multipl));
	
	g_syncLock.Aquire();
	
	g_tickMultiplier = multipl;
	g_dialTarget = target;

	if ( bUseSt == true ){
		g_dialValue = startVal;
	}

	g_syncLock.Release();
}


/******************************************************************************/
/*
/*  Funktion:		GetDialTarget
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		development purpose only!			
/*		
/******************************************************************************/
int Communication::GetDialTarget(void)
{	
	g_syncLock.Aquire();
	int val = g_dialTarget;
	g_syncLock.Release();
	return val;
}



/******************************************************************************/
/*
/*  Funktion:		RedirectDialRequest
/*	
/*
/*	Parameters:		--
/*
/*	Return Value:	--
/*
/*	Comments:		development purpose only!
/*		
/******************************************************************************/
void RedirectDialRequest(Communication *pComm, int direction, int ticks)
{	
	DBGTRACE(("RedirectDialRequest entered."));

	//nur dreh-Events auswerten
	if (direction == DIALLEFT ){
		DBGTRACE(("RedirectDialRequest detected left with %d ticks.", ticks));
		g_dialValue -= (ticks * g_tickMultiplier);
	}
	else if ( direction == DIALRIGHT ){
		DBGTRACE(("RedirectDialRequest detected right with %d ticks.", ticks));
		g_dialValue += (ticks * g_tickMultiplier);
	}
	else{
		DBGTRACE(("RedirectDialRequest unsupported action."));
		return;
	}

	//Target modifizieren
	switch( g_dialTarget ){

	case TARGETVOL:
		DBGTRACE(("RedirectDialRequest requesting volume %d ", g_dialValue));
		
		//einzustellender Wert innerhalb der Grenzen?
		if ( g_dialValue < g_pm_AudioCaps->volMin ){
			g_dialValue = g_pm_AudioCaps->volMin;
		}
		else if ( g_dialValue > g_pm_AudioCaps->volMax ){
			g_dialValue = g_pm_AudioCaps->volMax;
		}

		pComm->Audio_SetVolume(g_dialValue);
		break;

	case TARGETBAL:
		DBGTRACE(("RedirectDialRequest requesting balance %d ", g_dialValue));

		//einzustellender Wert innerhalb der Grenzen?
		if ( g_dialValue < g_pm_AudioCaps->balLeft ){
			g_dialValue = g_pm_AudioCaps->balLeft;
		}
		else if ( g_dialValue > g_pm_AudioCaps->balRight ){
			g_dialValue = g_pm_AudioCaps->balRight;
		}

		pComm->Audio_SetBalance(g_dialValue);
		break;

	case TARGETFAD:
		DBGTRACE(("RedirectDialRequest requesting fader %d ", g_dialValue));

		//einzustellender Wert innerhalb der Grenzen?
		if ( g_dialValue < g_pm_AudioCaps->fadFront ){
			g_dialValue = g_pm_AudioCaps->fadFront;
		}
		else if ( g_dialValue > g_pm_AudioCaps->fadRear ){
			g_dialValue = g_pm_AudioCaps->fadRear;
		}

		pComm->Audio_SetFader(g_dialValue);
		break;

	case TARGETFMEUR:
		DBGTRACE(("RedirectDialRequest requesting fm eur %d ", g_dialValue));
		
		//(EUROPE-1) muss gltigen Index fr Array liefern
		assert( (EUROPE > 0) && (EUROPE < 5) );
		
		//(FM-1) muss gltigen Index fr Array liefern
		assert( (FM > 0) && (FM < 6) );

		//einzustellender Wert innerhalb der Grenzen?
		if ( g_dialValue <  (int)g_areaDescr[(EUROPE-1)].band[(FM-1)].minFreq ){
			g_dialValue = g_areaDescr[(EUROPE-1)].band[(FM-1)].minFreq;
			
			DBGTRACE(("RedirectDialRequest fmeur lower limit %d hit.",
				g_areaDescr[(EUROPE-1)].band[(FM-1)].minFreq));
		}
		else if ( g_dialValue > (int)g_areaDescr[(EUROPE-1)].band[(FM-1)].maxFreq ){
			g_dialValue = g_areaDescr[(EUROPE-1)].band[(FM-1)].maxFreq;
			
			DBGTRACE(("RedirectDialRequest fmeur upper limit %d hit.",
				g_areaDescr[(EUROPE-1)].band[(FM-1)].maxFreq));
		}
		
		DBGTRACE(("RedirectDialRequest now requesting fm eur %d ", g_dialValue));
		pComm->Tuner_SetFrequency(g_dialValue);
		break;

	case TARGETFMUSA:
		DBGTRACE(("RedirectDialRequest requesting fm USA %d ", g_dialValue));

		//(USA-1) muss gltigen Index fr Array liefern
		assert( (USA > 0) && (USA < 5) );
		
		//(FM-1) muss gltigen Index fr Array liefern
		assert( (FM > 0) && (FM < 6) );

		//einzustellender Wert innerhalb der Grenzen?
		if ( g_dialValue <  (int)g_areaDescr[(USA-1)].band[(FM-1)].minFreq ){
			g_dialValue = (int)g_areaDescr[(USA-1)].band[(FM-1)].minFreq;
			
			DBGTRACE(("RedirectDialRequest fmusa lower limit %d hit.",
				g_areaDescr[(USA-1)].band[(FM-1)].minFreq));
		}
		else if ( g_dialValue > (int)g_areaDescr[(USA-1)].band[(FM-1)].maxFreq ){
			g_dialValue = (int)g_areaDescr[(USA-1)].band[(FM-1)].maxFreq;
			
			DBGTRACE(("RedirectDialRequest fmusa upper limit %d hit.",
				g_areaDescr[(USA-1)].band[(FM-1)].maxFreq));
		}
		
		pComm->Tuner_SetFrequency(g_dialValue);
		break;

	default: 
		return;
	}
}


/*************************************************************
/*
/* exclusive access to data
/*
/************************************************************/
DataLock::DataLock()
{
	InitializeCriticalSection(&m_lock);
}

DataLock::~DataLock()
{
	DeleteCriticalSection(&m_lock);
}

bool DataLock::Aquire(void)
{
	EnterCriticalSection(&m_lock);
	return true;
}

void DataLock::Release(void)
{
	LeaveCriticalSection(&m_lock);
}


/*************************************************************
/*
/* Debug-Tracing
/*
/************************************************************/
void printfToTraceDevice(LPSTR fmt, ...)
{	
	CHAR buf[256];
	va_list args;
	va_start(args, fmt);

	sprintf(buf, "COMM_IF: ");
	vsprintf(buf + strlen(buf), fmt, args);
	strcat(buf, "\r\n");
	OutputDebugString(buf);
}

