#ifndef __Communication_h__
#define __Communication_h__

#include "globals.h"
#include <assert.h>


//*****************************************************************************
//**** debug definition *******************************************************
//*****************************************************************************
#ifdef _DEBUG
	//Tracing on
	#define DBGTRACE(args) printfToTraceDevice args
#else
	//Tracing off
	#define DBGTRACE(args)
#endif

#define IN
#define OUT


//*****************************************************************************
//**** device identifier ******************************************************
//*****************************************************************************
#define CANMODUL	(1)
#define AUDIOMODUL	(2)
#define POWERMODUL  (4)
#define KLINEMODUL1	(5)
#define KLINEMODUL2	(6)

//*****************************************************************************
//**** class definition *******************************************************
//*****************************************************************************
class Communication
{
public:
	
	//Constructor, Destructor
	Communication(FN_CALLBACK fnName);
	~Communication();
	
	//Connect, Disconnect
	bool Connect(void);
	bool DisConnect(void);



	/******************** Audio functions ********************/
	bool Audio_RequestCaps(void);
	
	void Audio_ViewCaps(OUT AUDIOCAPS *caps);
	
	bool Audio_SetVolume(IN int value);
	
	bool Audio_SetBass(IN int value);

	bool Audio_SetTreble(IN int value);
	
	bool Audio_SetLoudness(IN int value);
	
	bool Audio_SetBalance(IN int value);
	
	bool Audio_SetFader(IN int value);
	
	void Audio_Mute(IN bool value);

	bool Audio_SelectSource(IN int value);



	/******************* Tuner Functions *********************/
	bool Tuner_RequestCaps(void);

	void Tuner_ViewCaps(OUT AREADESCR *caps, IN int area);
	
	void Tuner_GetFrequencyParams(OUT FREQPARAMS *params);

	bool Tuner_ScanBand(IN int minFreq, IN int maxFreq, IN int step);

	bool Tuner_SetFrequency(IN int value);

	bool Tuner_SelectBand(IN int value);

	bool Tuner_SelectArea(IN int value);

	void Tuner_GetScanResult(OUT FREQULIST *list);



	/******************* CANSJA Functions *********************/
	bool CAN_PhysicalConnect(
		IN int		physConId, 
		IN int		conType, 
		IN int		busSpeed
	);

	DWORD CAN_LogicalConnect(
		IN DWORD	logicalConId,
		IN DWORD	acceptanceMask, 
		IN DWORD	acceptanceCode,
		IN BYTE		connectionflags,
		IN BYTE		mailboxFlags
	);

	bool CAN_SendMessage(
		IN DWORD			CANId,
		IN BOOL				ack,
		IN BOOL				rtr,
		IN BOOL				conf,
		IN unsigned char	*pCANMsg, 
		IN int				CANMsgLen
	);
	
	void CAN_ViewConnectionInfo(
		OUT CANINFO		*canInfo
	);

	bool CAN_PhysicalDisconnect(void);
	
	DWORD CAN_LogicalDisconnect(void);



	/******************* K-Line Functions **********************/
	bool KLine_QueryVersion(
		void
	);
	
	void KLine_ViewConnectionInfo(
		IN	int				kLineChannel, 
		OUT KLINEINFO_P		kLineInfo
	);

	bool KLine_RequestCapabilities(
		IN	int				kLineChannel
	);

	bool KLine_PhysicalConnect(
		IN	int				kLineChannel,  
		IN	int				physConId
	);

	bool KLine_LogicalConnect(
		IN	int				kLineChannel, 
		IN	int				physConId, 
		IN	int				logConId, 
		IN	int				protocolID, 
		IN	int				bitRate, 
		IN	int				nodeID
	);

	bool KLine_SendData(
		IN	int				kLineChannel, 
		IN	int				physConId, 
		IN	int				logConId, 
		IN	char			blockId,
		IN	char			flags,
		IN	char*			pData, 
		IN	DWORD			dataLen
	);

	bool KLine_PhysicalDisconnect(
		IN	int				kLineChannel, 
		IN	int				physConId
	);

	bool KLine_LogicalDisconnect(
		IN	int				kLineChannel, 
		IN	int				physConId, 
		IN	int				logConId
	);



	/******************** Power Management *********************/
	bool Power_SwitchOffSupply(IN unsigned int delayMilliSeconds);

	bool Power_QueryTerm15(void);

	bool Power_QueryCapacity(void);

	bool SendPCStartUpMessage(void);

	

	/******************* other Functions ***********************/
	bool QueryCurrentValue(unsigned int nItem);

	bool SendUserDefinedMessage(
		IN unsigned char*	pMessage, 
		IN unsigned int		len
	);

	void SetDialTarget(
		IN int target, 
		IN int startVal, 
		IN int tickMultiplier,
		IN bool bUseStartVal
	);

	int GetDialTarget(void);


	//Inlines
	inline void SetPriority(unsigned char value){
		m_CurrentPrio = value;
	};

	inline unsigned char GetPriority(void){
		return m_CurrentPrio;
	};

	inline unsigned long GetLinkId(void){
		return m_linkId;
	};


private:

	bool SendWithPrio(unsigned char* pMessage);	
	int	 FindKLineChannel(int modulID);

	AUDIOCAPS		m_AudioCaps;	
	CANINFO			m_CANInfo;
	KLINEINFO		m_KLineInfo[2];

	bool			m_connected;
	unsigned char	m_CurrentPrio;
	unsigned long	m_linkId;
	
	FN_CALLBACK		m_Callback;
};



//*****************************************************************************
//**** exclusive access to data ***********************************************
//*****************************************************************************
class DataLock
{
public:
	DataLock();
	~DataLock();

	bool Aquire();
	void Release();
private:
	CRITICAL_SECTION    m_lock;
};


//*****************************************************************************
//**** Debug-Tracing **********************************************************
//*****************************************************************************
void printfToTraceDevice(LPSTR fmt, ...);

#endif // __Communication_h__

